/*--------------------------- MegaWave2 Module -----------------------------*/
/* mwcommand 
  name = {fgrain}; 
  version = {"1.2"}; 
  author = {"Pascal Monasse, Frederic Guichard, G. Facciolo"}; 
  function = {"Grain filter of an image"}; 
  usage = { 
    'a': [min_area=20]-> pMinArea   "Min area of grains we keep", 
    image_in -> pFloatImageInput    "Input fimage", 
    image_out <- pFloatImageOutput  "Output fimage" 
    }; 
*/ 
/*----------------------------------------------------------------------
 v1.2 (04/2007): simplified header (LM)
 v1.3 (2013): portable version (GF)
----------------------------------------------------------------------*/

#include "mw3.h" 
#include "iio.h"
#include "stdio.h"
#include "stdlib.h"

extern void flst();
extern void flst_reconstruct();

//  Main function and input/output.

Fimage fimageread(char* name)
{
    int nx,ny,nc;
    float *fdisp = iio_read_image_float_split(name, &nx, &ny, &nc);
    Fimage out = mw_change_fimage(NULL,ny,nx);
    free(out->gray);
    out->gray=fdisp;
    return out;
}

void fimagewrite(char* name, Fimage i)
{
    iio_save_image_float_split(name, i->gray, i->ncol, i->nrow, 1);
}


/* This removes the shapes from the tree associated to pFloatImageInput
that are too small (threshold *pMinArea). As a consequence all the remaining 
shapes of pFloatImageOutput are of area larger or equal than *pMinArea */ 

/*
 Modified by Charles Hessel on Wednesday, 2015-09-02
 just to allow this grain filter to output several filtered images with different
 minimum grain areas.
 */
void mw_fgrain(Fimage pFloatImageInput, Fimage pFloatImageOutput, int Nb_grain, int *p_grain_areas, char *p_output_names)
{ 
    int i=0,n=0;
    char tmp_name[13] = {'g','r','a','i','n','9','9','9','.','p','n','g','\0'};
    Shapes pTree;
    
    if(mw_change_fimage(pFloatImageOutput, pFloatImageInput->nrow,
                        pFloatImageInput->ncol) == NULL)
        mwerror(FATAL, 1,
                "fgrain --> Not enough memory to allocate the output image");
    if((pTree = mw_new_shapes()) == NULL)
        mwerror(FATAL, 1,
                "fgrain --> Not enough memory to allocate the tree of shapes");
    
    /* Compute the Level Sets Transform of the input image */
    flst(NULL, pFloatImageInput, pTree, NULL, NULL);
    
    // Debug
    // printf("Nous sommes bien dans mw_grain, la flst vient de se terminer.\n");
    
    for(n=0;n<Nb_grain;n++) {
        /* Kill too small grains.
         Bound i>0 because it is forbidden to delete the root, at index 0 */
        for(i = pTree->nb_shapes-1; i > 0; i--)
            if(pTree->the_shapes[i].area < *(p_grain_areas+n))
                pTree->the_shapes[i].removed = (char)1;
        
        /* Reconstruct in pFloatImageOutput the modified tree of shapes */
        flst_reconstruct(pTree, pFloatImageOutput);
        
        for(i=0;i<13;i++) tmp_name[i] = *(p_output_names+i+n*13);
        
        // Debug
        // printf("La prochaine instruction est d enregistrer l'image %s\n",tmp_name);
        
        fimagewrite(tmp_name, pFloatImageOutput);
        
    }
    mw_delete_shapes(pTree);
} 



/*in and out must be allocated*/
/*
 fgrain:
 Modified by Charles Hessel on Wednesday, 2015-09-02
 For "Multigrain Filter"... (see the description for "main")
 */
void fgrain(float *in, int nx, int ny, float *out, int Nb_grain, int *p_grain_areas, char *p_output_names) {
    int i;
    Fimage mwin = mw_change_fimage(NULL,ny,nx);
    Fimage mwout = mw_new_fimage();
    for(i=0;i<nx*ny;i++) mwin->gray[i] = in[i];
    
    // Debug
    /*
     printf("Nous sommes bien dans fgrain.\n");
     printf("Nb_grain = %d\n",Nb_grain);
     printf("Grain Areas = ");
     for(i=0;i<Nb_grain;i++) printf("%d ",p_grain_areas[i]);
     printf("\n");
    */
    
    mw_fgrain(mwin, mwout, Nb_grain, p_grain_areas, p_output_names);
    
    //for(i=0;i<nx*ny;i++) out[i] = mwout->gray[i];
    mw_delete_fimage(mwin);
    mw_delete_fimage(mwout);
}


/*
 main:
 Modifications by Charles Hessel on Wednesday, 2015-09-02
 For a "Multi-Grain Filter" able to output several images filtered with
 N different grain sizes. N being set by user, the grain sizes too.
 */

int main(int argc, char* argv[]) 
{
    // default parameters
    // If ./multigrain then help
    // If ./multigrain img_in then N=1 and t1=20
    // If ./multigrain img_in N then error and help
    // If ./multigrain img_in N t1, t2, ... but not N tresholds, then error.
    
    int Nb_grain=1;         // default value for the number of output images
    int n=0,i=0;            // iterator
    int *p_grain_areas;     // table for the N areas (N thresholds)
    char *p_output_names;   // table for the N names of the output images
    char tmp_name[13] = {'g','r','a','i','n','0','0','0','.','p','n','g','\0'};
                            // temporary char table for output image names
    
    if(argc==1) {
        fprintf(stderr, "Multi-Grain filter of an image\n");
        fprintf(stderr, "syntax: %s image_in [N=1 t1=20]\n", argv[0]);
        fprintf(stderr, "    or: %s image_in N t1 t2 t3 ... tN\n", argv[0]);
        return 1;
    }
    if(argc==2) {
        p_grain_areas = (int*) malloc(Nb_grain*sizeof(int));
        p_output_names = (char*) malloc(Nb_grain*13*sizeof(char));
        // Defualt values:
        *p_grain_areas = 20;
        for(i=0;i<13;i++) *(p_output_names+i) = tmp_name[i];
    }
    if(argc==3) {
        fprintf(stderr, "Multi-Grain filter of an image\n");
        fprintf(stderr, "you must give the minimum area for each of the %s output images\n", argv[3]);
        fprintf(stderr, "syntax: %s image_in N t1 t2 ... tN\n", argv[0]);
        return 1;
    }
    if(argc>=4) {
        Nb_grain=atoi(argv[2]);
        // Debug
        // printf("There will be %d output images\n",Nb_grain);
        if(argc < 3+Nb_grain) {
            fprintf(stderr, "not enough areas. There are %d areas and it should be %d\n",argc,Nb_grain);
            return 1;
        }
        p_grain_areas = (int*) malloc(Nb_grain*sizeof(int));
        p_output_names = (char*) malloc(Nb_grain*13*sizeof(char));
        
        for(n=0; n<Nb_grain; n++) {
            p_grain_areas[n] = atoi(argv[3+n]);
            
            sprintf(tmp_name,"grain%03d.png",n);
            
            // Debug
            // printf("%s\n",tmp_name);
            
            for(i=0;i<13;i++) p_output_names[13*n+ i] = tmp_name[i];
            
        }
        
        // Debug
        /*
        for(n=0;n<Nb_grain;n++) {
            for(i=0;i<13;i++) {
                printf("%c",*(p_output_names+i+n*13));
            }
            printf(" ");
        }
        printf("\n");
        */
    }
    
    // Debug
    // for(n=0;n<Nb_grain;n++) {
    //     printf("Area %d/%d : %d\n",n+1,Nb_grain,p_grain_areas[n]);
    // }
    
    
    Fimage in = fimageread(argv[1]);
    Fimage out = mw_change_fimage(NULL,in->nrow,in->ncol);
    fgrain(in->gray, in->ncol, in->nrow, out->gray, Nb_grain, p_grain_areas, p_output_names);
    
    return 0;
}
